//
// Accounts or entries into the hiscore list. 
// An account holds all data needed to rank a player, and some extra.
//
// Warning: The Account struct/class doesn't have any constructors/destructor, so all initialization and 
//          linking is done manually. So malloc()/free() is used to indicate this. Sadly can lead to harder 
//          to read code and add bugs in the future.
//

#ifndef ACCOUNT_H
#define ACCOUNT_H

struct Account
{
	unsigned char		bMode;								// Mode: Normal, Master, Plus or Death
	char				aName[3];							// Name with 3 characters only, and no NULL-terminator
	char				aCode[16];							// Code with 16 characters only, and no NULL-terminator
	unsigned short		wGrade;								// Visible grade
	unsigned short		wRankedLine;						// Ranked line for finished games
	unsigned long		dwTime;								// Mastering time
	unsigned long		dwScore;							// Score reached
	unsigned short		wLevel;								// Level reached
	unsigned char		abMedallions[6];					// 6 medallions

	char *				pszComment;							// Pointer to an allocated buffer holding the variable width string

	Account *			pNext;								// Next account in the list or NULL to indicate the end
};

extern Account * pAccountNormalRoot;						// Root for normal mode hiscore list
extern Account * pAccountMasterRoot;						// Root for master mode hiscore list
extern Account * pAccountPlusRoot;							// Root for plus mode hiscore list
extern Account * pAccountDeathRoot;							// Root for death mode hiscore list

// So we know how much to scroll the window
extern unsigned int uNumNormalAccounts;
extern unsigned int uNumMasterAccounts;
extern unsigned int uNumPlusAccounts;
extern unsigned int uNumDeathAccounts;

void BuildRankingPriority();

bool InsertAccount(Account ** ppAccountFrom, Account * pAccountNew);

//
// At this point in time we have no reason to remove one account entry at a time. So we don't include any function such as DeleteAccount()
//

bool AddAccount(Account * pAccountNew);

void DeleteAccounts(Account ** ppAccountRoot);

void CleanupAccounts();

void AddModeHeaders();

#endif
