//
// Awesome Scrollbar by LOst
//
// Copyright (C) 2014 LOst
//
// Yep, text mode scrollbar with a proportional elevator evenly aligned and rounded to text blocks. 
//

#include "AwesomeScrollbar.h"

// So we can paint stuff
#include "ConsoleIO.h"

AwesomeScrollbar::AwesomeScrollbar(
					AwesomeScrollbar::ScrollType Type, 
					ShortRect const & rcCover, 
					signed short Min, 
					signed short Max, 
					signed short Page, 
					signed short Pos)
:	Type(Type), 
	rc(rcCover), 
	Min(Min), 
	Max(Max), 
	Page(Page), 
	Pos(Pos), 

	bForeColor(15), 
	bBkgndColor(0), 

	bCharArrowUp(24), 
	bCharArrowDown(25), 
	bCharArrowLeft(27), 
	bCharArrowRight(26), 
	bCharTrack(0xB0), 
	bCharElevator(0xB2)
{
	this->UpdateScrollbar();
}

void AwesomeScrollbar::UpdateScrollbar()
{
	signed short MaxPage = ((this->Max - this->Min) + 1);

	if (this->Page >= MaxPage)
	{
		this->Calc.FEnabled = false;
		this->Pos = 0;
	}
	else
	{
		this->Calc.FEnabled = true;
	}

	signed short MaxPos = (this->Max - (this->Page == 0 ? 0 : this->Page - 1));

	// Limit position within minimum and maximum values
	if (this->Pos < this->Min)
	{
		this->Pos = this->Min;
	}
	else if (this->Pos > MaxPos)
	{
		this->Pos = MaxPos;
	}

	// Calculate position of the scrollbar for painting
	if (this->Type == AwesomeScrollbar::ScrollH)
	{
		this->Calc.At = this->rc.top;
		this->Calc.Start = this->rc.left;
		this->Calc.End = this->rc.right;
	}
	else
	{
		this->Calc.At = this->rc.left;
		this->Calc.Start = this->rc.top;
		this->Calc.End = this->rc.bottom;
	}

	this->Calc.Min = this->Calc.Start + 1;
	this->Calc.Max = this->Calc.End - 1;

	// Find the size and position of the elevator (max page)
	this->Calc.ScrollRange = ((this->Max - this->Min) + 1);

	if (this->Page != 0)
	{
		// We use rounding here
		this->Calc.ElevatorSize = (((this->Calc.Max - this->Calc.Min) * this->Page) +		// Mul
								   (this->Calc.ScrollRange >> 1)) /							// Round to nearest
								  this->Calc.ScrollRange;									// Div

		if (this->Calc.ElevatorSize < 1)
		{
			this->Calc.ElevatorSize = 1;
		}
	}
	else
	{
		this->Calc.ElevatorSize = 1;
	}

	this->Calc.NonElevator = this->Calc.Max - this->Calc.ElevatorSize - this->Calc.Min;

	signed short PageRange = (this->Calc.ScrollRange - (this->Page == 0 ? 1 : this->Page));

	if (PageRange == 0)
	{
		this->Calc.ElevatorLo = this->Calc.Min;
	}
	else
	{
		// We use rounding here
		this->Calc.ElevatorLo = this->Calc.Min + 
								((((this->Pos - this->Min) * this->Calc.NonElevator) +		// Mul
								 (PageRange >> 1)) /										// Round to nearest
								 PageRange);												// Div
	}

	this->Calc.ElevatorHi = this->Calc.ElevatorLo + this->Calc.ElevatorSize;
}

bool AwesomeScrollbar::PaintScrollbar()
{
	if (this->Calc.FEnabled != true)
	{
		// The scrollbar isn't needed, so give the caller a chance to draw things in place of the scrollbar
		return false;
	}

	ShortRect rc;

	if (this->Type == AwesomeScrollbar::ScrollH)
	{
		// Draw lower track
		rc.top = this->Calc.At;
		rc.bottom = rc.top + 1;

		rc.left = this->Calc.Min;
		rc.right = this->Calc.ElevatorLo;

		PaintRect(rc, this->bCharTrack, this->bForeColor, this->bBkgndColor);

		// Draw upper track
		rc.left = this->Calc.ElevatorHi;
		rc.right = this->Calc.Max;

		PaintRect(rc, this->bCharTrack, this->bForeColor, this->bBkgndColor);

		// Draw arrow buttons
		PaintChar(this->Calc.Start, this->Calc.At, this->bCharArrowLeft, this->bBkgndColor, this->bForeColor);
		PaintChar(this->Calc.End - 1, this->Calc.At, this->bCharArrowRight, this->bBkgndColor, this->bForeColor);

		// Draw elevator thumb
		rc.left = this->Calc.ElevatorLo;
		rc.right = this->Calc.ElevatorHi;

		PaintRect(rc, this->bCharElevator, this->bForeColor, this->bBkgndColor);
	}
	else
	{
		// Draw lower track
		rc.left = this->Calc.At;
		rc.right = rc.left + 1;

		rc.top = this->Calc.Min;
		rc.bottom = this->Calc.ElevatorLo;

		PaintRect(rc, this->bCharTrack, this->bForeColor, this->bBkgndColor);

		// Draw upper track
		rc.top = this->Calc.ElevatorHi;
		rc.bottom = this->Calc.Max;

		PaintRect(rc, this->bCharTrack, this->bForeColor, this->bBkgndColor);

		// Draw arrow buttons
		PaintChar(this->Calc.At, this->Calc.Start, this->bCharArrowUp, this->bBkgndColor, this->bForeColor);
		PaintChar(this->Calc.At, this->Calc.End - 1, this->bCharArrowDown, this->bBkgndColor, this->bForeColor);

		// Draw elevator thumb
		rc.top = this->Calc.ElevatorLo;
		rc.bottom = this->Calc.ElevatorHi;

		PaintRect(rc, this->bCharElevator, this->bForeColor, this->bBkgndColor);
	}

	return true;
}

AwesomeScrollbar::ScrollHT AwesomeScrollbar::ScrollHitTest(signed short x, signed short y)
{
	bool FInside;
	signed short * z;

	if (this->Type == ScrollH)
	{
		FInside = (y == this->Calc.At);
		z = &x;
	}
	else
	{
		FInside = (x == this->Calc.At);
		z = &y;
	}

	if (FInside == true)
	{
		if (*z == this->Calc.Start)
		{
			return ScrollHTArrowLo;
		}
		else if (*z == (this->Calc.End - 1))
		{
			return ScrollHTArrowHi;
		}

		if (*z >= this->Calc.Min && *z < this->Calc.Max)
		{
			if (*z < this->Calc.ElevatorLo)
			{
				return ScrollHTTrackLo;
			}
			else if (*z >= this->Calc.ElevatorHi)
			{
				return ScrollHTTrackHi;
			}

			return AcrollHTElevator;
		}
	}

	return ScrollHTNot;
}

bool AwesomeScrollbar::TrackElevator(signed short CalcPos)
{
	if (this->Calc.FEnabled != true)
	{
		// The scrollbar isn't enabled
		return false;
	}

	signed short OldPos = this->Pos;

	// Find the new pos from the elevator's position. We use rounding again
	if (CalcPos < this->Calc.Min)
	{
		this->TrackPos = this->Min;
	}
	else if (CalcPos >= (this->Calc.Min + this->Calc.NonElevator))
	{
		this->TrackPos = (this->Max - (this->Page == 0 ? 0 : this->Page - 1));
	}
	else
	{
		if (this->Calc.NonElevator == 0)
		{
			this->TrackPos = this->Min;
		}
		else
		{
			// We use rounding here
			this->TrackPos = this->Min + 
							 ((((this->Max - this->Min - (this->Page == 0 ? 0 : this->Page - 1)) * 
								(CalcPos - this->Calc.Min)) +						// Mul
							   (this->Calc.NonElevator >> 1)) /						// Round to nearest
							  this->Calc.NonElevator);								// Div
		}
	}

	return (this->TrackPos != OldPos);
}
