//
// Awesome Scrollbar by LOst
//
// Copyright (C) 2014 LOst
//
// Yep, text mode scrollbar with a proportional elevator evenly aligned and rounded to text blocks. 
//

#ifndef AWESOME_SCROLLBAR_H
#define AWESOME_SCROLLBAR_H

#include "ShortArea.h"

class AwesomeScrollbar
{
public:
	enum ScrollType
	{
		ScrollH, 
		ScrollV
	};

	enum ScrollHT
	{
		ScrollHTNot, 
		ScrollHTArrowLo, 
		ScrollHTArrowHi, 
		ScrollHTTrackLo, 
		ScrollHTTrackHi, 
		AcrollHTElevator
	};

public:
	ScrollType		Type;				// Type of scrollbar
	ShortRect		rc;					// Position and size of the scrollbar

	// Scroll range is Max-Page-Min+1.
	// Allowed scroll is between Min and (Max - Page). When Page > (Max - Min) then it is disabled or hidden AND 
	// Page is set to (Max - Min) and Pos is set to Min. The size of the elevator cannot be less than 1 char when 
	// page gets too small, and the elevator disappears when the scrollbar is too short!
	signed short	Min;				// Min scroll position
	signed short	Max;				// Max scroll position
	signed short	Page;				// Visible scroll area
	signed short	Pos;				// Position of the elevator
	signed short	TrackPos;			// Current position while dragging the elevator

	unsigned char	bForeColor;			// Foreground and background colors of the scrollbar
	unsigned char	bBkgndColor;

	unsigned char	bCharArrowUp;		// Text themed scrollbar elements
	unsigned char	bCharArrowDown;
	unsigned char	bCharArrowLeft;
	unsigned char	bCharArrowRight;
	unsigned char	bCharTrack;
	unsigned char	bCharElevator;

private:
	struct ScrollBarCalc
	{
		bool			FEnabled;		// Enabled if scroll is needed
		signed short	At;				// Coordiante where scrollbar is located
		signed short	Start;			// Coordinate of scrollbar start
		signed short	End;			// Coordinate of scrollbar end
		signed short	Min;			// Minimum position
		signed short	ElevatorLo;		// Top or left of scrollbar elevator
		signed short	ElevatorHi;		// Bottom or right of scrollbar elevator
		signed short	Max;			// Maximum position
		signed short	ScrollRange;	// ((Max - Min) + 1)
		signed short	ElevatorSize;	// Size of the elevator
		signed short	NonElevator;	// Size of non elevator
	};

	ScrollBarCalc	Calc;				// Calculation of scrollbar stuff

public:
	AwesomeScrollbar(
		ScrollType Type, 
		ShortRect const & rcCover, 
		signed short Min, 
		signed short Max, 
		signed short Page, 
		signed short Pos);

public:
	void UpdateScrollbar();

	bool PaintScrollbar();

	ScrollHT ScrollHitTest(signed short x, signed short y);

	bool TrackElevator(signed short CalcPos);

	inline signed short GetElevatorGrip(signed short At) const
	{
		return At - this->Calc.ElevatorLo;
	}
};

#endif
