#include "ConsoleIO.h"

#include <Windows.h>
#include <stdio.h>

HANDLE hStdOutOld			= INVALID_HANDLE_VALUE;
HANDLE hStdOutNew			= INVALID_HANDLE_VALUE;
DWORD dwConsoleModeInOld;
CONSOLE_CURSOR_INFO CursorOld;
CONSOLE_CURSOR_INFO CursorNew;

bool SetupMaxConsoleScreenMode()
{
#if SMALL_FONT
	typedef BOOL (__stdcall * PFNSetConsoleFont)(HANDLE, DWORD);
	typedef COORD (__stdcall * PFNGetConsoleFontSize)(HANDLE hConsoleOutput, DWORD nFont);
	typedef DWORD (__stdcall * PFNGetNumberOfConsoleFonts)();

	PFNSetConsoleFont pSetConsoleFont					= NULL;
	PFNGetConsoleFontSize pGetConsoleFontSize			= NULL;
	PFNGetNumberOfConsoleFonts pGetNumberOfConsoleFonts	= NULL;
	
	HMODULE hMod = GetModuleHandleA("KERNEL32.DLL");

	if (hMod == NULL)
	{
		return false;
	}

	pGetNumberOfConsoleFonts = reinterpret_cast<PFNGetNumberOfConsoleFonts>(GetProcAddress(hMod, "GetNumberOfConsoleFonts"));
	pGetConsoleFontSize		= reinterpret_cast<PFNGetConsoleFontSize>(GetProcAddress(hMod, "GetConsoleFontSize"));
	pSetConsoleFont			= reinterpret_cast<PFNSetConsoleFont>(GetProcAddress(hMod, "SetConsoleFont"));
#endif

	bool FRet = false;

	do
	{
		hStdOutOld = GetStdHandle(STD_OUTPUT_HANDLE);

		if (hStdOutOld != INVALID_HANDLE_VALUE)
		{
			GetConsoleCursorInfo(hStdOutOld, &CursorOld);
		}
		else
		{
			break;
		}

		HANDLE hStdIn = GetStdHandle(STD_INPUT_HANDLE);

		if (hStdIn == INVALID_HANDLE_VALUE)
		{
			break;
		}

		if (GetConsoleMode(hStdIn, &dwConsoleModeInOld) == 0)
		{
			break;
		}
		
		if (SetConsoleMode(hStdIn, (ENABLE_WINDOW_INPUT | 
									ENABLE_MOUSE_INPUT)) == 0)
		{
			break;
		}

		hStdOutNew = CreateConsoleScreenBuffer(GENERIC_READ | GENERIC_WRITE, 
											   0, 
											   NULL, 
											   CONSOLE_TEXTMODE_BUFFER, 
											   NULL);

		if (hStdOutNew == INVALID_HANDLE_VALUE)
		{
			break;
		}

		CursorNew = CursorOld;
		CursorNew.bVisible = FALSE;

		SetConsoleCursorInfo(hStdOutNew, &CursorNew);

		COORD coord = {(unsigned short)(COLUMNS), (unsigned short)(ROWS)};
		SMALL_RECT sm = {0, 0, coord.X - 1, coord.Y - 1};

		SetConsoleActiveScreenBuffer(hStdOutNew);

		SetStdHandle(STD_OUTPUT_HANDLE, hStdOutNew);

		SetConsoleScreenBufferSize(hStdOutNew, coord);

		SetConsoleWindowInfo (hStdOutNew, TRUE, &sm);

#if SMALL_FONT
		if (pGetNumberOfConsoleFonts != NULL && 
			pGetConsoleFontSize != NULL && 
			pSetConsoleFont != NULL)
		{
			DWORD dwNumConsoleFonts = (*pGetNumberOfConsoleFonts)();

			for (DWORD dwIndex = 0; dwIndex < dwNumConsoleFonts; dwIndex++)
			{
				coord = (*pGetConsoleFontSize)(hStdOutNew, dwIndex);

				if (coord.X == FONT_X && 
					coord.Y == FONT_Y)
				{
					(*pSetConsoleFont)(hStdOutNew, dwIndex);
					break;
				}
			}
		}
#endif

		FRet = true;
	}
	while (false);

	if (FRet == false)
	{
		RevertToDefaultScreenMode ();
	}

	return FRet;
}

void RevertToDefaultScreenMode()
{
	// Make sure the color attributes are set to default
	PaintChar(COLUMNS, ROWS, ' ', 7, 15);

	if (hStdOutOld != INVALID_HANDLE_VALUE)
	{
		SetStdHandle(STD_OUTPUT_HANDLE, hStdOutOld);

		SetConsoleCursorInfo(hStdOutOld, &CursorOld);

		SetConsoleActiveScreenBuffer(hStdOutOld);

		hStdOutOld = INVALID_HANDLE_VALUE;
	}

	if (hStdOutNew == INVALID_HANDLE_VALUE)
	{
		CloseHandle(hStdOutNew);
		hStdOutNew = INVALID_HANDLE_VALUE;
	}

	if (dwConsoleModeInOld != 0)
	{
		HANDLE hStdIn = GetStdHandle(STD_INPUT_HANDLE);

		if (hStdIn != INVALID_HANDLE_VALUE)
		{
			SetConsoleMode(hStdIn, dwConsoleModeInOld);
		}
	}
}

void ConsoleOut(signed short x, signed short y, char * pszText, signed short len, char ForeColor, char BkgndColor)
{
	if (pszText == NULL)
	{
		return;
	}

	COORD coord = {(unsigned short)x, (unsigned short)y};

	unsigned int uRealLen = (unsigned int)strlen(pszText);

	if ((unsigned int)len >= uRealLen)
	{
		len = uRealLen;
	}

	SetConsoleCursorPosition(hStdOutNew, coord);

	SetConsoleTextAttribute(hStdOutNew, ((ForeColor & 0xF) | ((BkgndColor & 0xF) << 4)));

	DWORD NumberOfCharsWritten;
	WriteConsoleA(hStdOutNew, pszText, len, &NumberOfCharsWritten, NULL);
}

void PaintChar(signed short x, signed short y, unsigned char bChar, char ForeColor, char BkgndColor)
{
	COORD coord = {(unsigned short)x, (unsigned short)y};

	SetConsoleCursorPosition(hStdOutNew, coord);

	SetConsoleTextAttribute(hStdOutNew, ((ForeColor & 0xF) | ((BkgndColor & 0xF) << 4)));

	DWORD NumberOfCharsWritten;
	WriteConsoleA(hStdOutNew, &bChar, 1, &NumberOfCharsWritten, NULL);
}

void PaintRect(const ShortRect & src, unsigned char bFill, char ForeColor, char BkgndColor)
{
	COORD coord = {(unsigned short)src.left, (unsigned short)src.top};
	signed short steps = src.bottom - src.top;
	signed short len = src.right - src.left;

	if (steps < 0 || len < 0)
	{
		return;
	}

	DWORD dwOut;

	for (signed short i = 0; i < steps; i++)
	{
		FillConsoleOutputCharacterA(hStdOutNew, bFill, len, coord, &dwOut);
		FillConsoleOutputAttribute(hStdOutNew, ((ForeColor & 0xF) | ((BkgndColor & 0xF) << 4)), len, coord, &dwOut);
		++coord.Y;
	}
}

void ScrollWindow(const ShortRect * psrcScroll, const ShortRect * psrcClip, signed short DestX, signed short DestY, unsigned char bFill, char ForeColor, char BkgndColor)
{
	SMALL_RECT ScrollRect, ClipRect;

	if (psrcScroll != NULL)
	{
		ScrollRect = *(SMALL_RECT *)psrcScroll;
		--ScrollRect.Right;
		--ScrollRect.Bottom;
		psrcScroll = (const ShortRect *)&ScrollRect;
	}

	if (psrcClip != NULL)
	{
		ClipRect = *(SMALL_RECT *)psrcClip;
		--ClipRect.Right;
		--ClipRect.Bottom;
		psrcClip = (const ShortRect *)&ClipRect;
	}

	COORD dwScrollTo;
	dwScrollTo.X = DestX;
	dwScrollTo.Y = DestY;

	CHAR_INFO chiFill = 
	{
		bFill, 
		((ForeColor & 0xF) | ((BkgndColor & 0xF) << 4))
	};

	ScrollConsoleScreenBuffer(
		hStdOutNew, 
		(SMALL_RECT *)psrcScroll, 
		(SMALL_RECT *)psrcClip, 
		dwScrollTo, 
		&chiFill);
}
