/************************************************
                      LTAP

           Score Validation Computation
                     by LOst

 Copyright (C) 2008 - 2014 LOst. 

 Attention:
   This code is NOT the "full version". 
   The encode function has been removed. If you 
   wish to use both the encode and decode 
   functions within your own code, please 
   contact me by PM.
   It's allowed to release this source to any of 
   many TGM clone developers to help expand the 
   TGM universe. 
   If you use, please give credits to me: LOst.
*************************************************/

#include "ScoreValidationDec.h"

#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

//////////////////////////////////////////////////
const char caValidationLetters[20] = 
{
	'3', '4', '6', '7', '9', 'C', 'D', 'E', 'F', 'G', 
	'H', 'J', 'K', 'M', 'P', 'R', 'T', 'W', 'X', 'Y' 
};
//////////////////////////////////////////////////
const char caHiscoreNameLetters[43] = 
{
	'&',	// Used for Doubles Mode's hiscore

	'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 
	'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 
	'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 
	'Y', 'Z', '0', '1', '2', '3', '4', '5', 
	'6', '7', '8', '9', '!', '?', '.', ' ', 

	'\\',	// R/D
	'['		// E/D
};
//////////////////////////////////////////////////
const char caScramble[256] = 
{
	0x03, 0x0A, 0x12, 0x01, 0x10, 0x0E, 0x08, 0x06, 0x0E, 0x01, 0x0F, 0x0E, 0x02, 0x03, 0x02, 0x08, 
	0x05, 0x10, 0x05, 0x07, 0x01, 0x0B, 0x12, 0x11, 0x0F, 0x0E, 0x09, 0x0F, 0x0A, 0x02, 0x05, 0x0B, 
	0x12, 0x0F, 0x07, 0x05, 0x0B, 0x09, 0x06, 0x0F, 0x0C, 0x10, 0x04, 0x03, 0x05, 0x0A, 0x0E, 0x02, 
	0x01, 0x10, 0x0E, 0x13, 0x07, 0x00, 0x05, 0x02, 0x12, 0x09, 0x11, 0x0C, 0x06, 0x10, 0x11, 0x04, 
	0x12, 0x00, 0x09, 0x12, 0x08, 0x09, 0x11, 0x05, 0x08, 0x09, 0x13, 0x02, 0x0D, 0x0A, 0x00, 0x0D, 
	0x12, 0x00, 0x0C, 0x12, 0x08, 0x0B, 0x0A, 0x13, 0x01, 0x06, 0x09, 0x0D, 0x0A, 0x02, 0x00, 0x0E, 
	0x03, 0x11, 0x04, 0x13, 0x06, 0x05, 0x0B, 0x07, 0x00, 0x05, 0x01, 0x05, 0x0A, 0x02, 0x13, 0x0E, 
	0x0F, 0x03, 0x06, 0x01, 0x0A, 0x00, 0x0B, 0x00, 0x02, 0x01, 0x0B, 0x0C, 0x07, 0x0D, 0x03, 0x09, 
	0x08, 0x0C, 0x11, 0x04, 0x0A, 0x12, 0x06, 0x07, 0x01, 0x07, 0x10, 0x12, 0x10, 0x0E, 0x06, 0x10, 
	0x0C, 0x0D, 0x0C, 0x08, 0x0C, 0x0A, 0x10, 0x11, 0x10, 0x04, 0x02, 0x03, 0x0B, 0x0D, 0x0C, 0x13, 
	0x07, 0x0D, 0x04, 0x05, 0x0F, 0x0C, 0x0F, 0x06, 0x12, 0x0B, 0x0D, 0x06, 0x13, 0x08, 0x0D, 0x0F, 
	0x11, 0x03, 0x00, 0x0E, 0x11, 0x0A, 0x01, 0x07, 0x0C, 0x06, 0x09, 0x04, 0x01, 0x11, 0x07, 0x04, 
	0x03, 0x08, 0x10, 0x0F, 0x00, 0x09, 0x04, 0x0D, 0x11, 0x12, 0x13, 0x03, 0x05, 0x0E, 0x02, 0x0B, 
	0x13, 0x0D, 0x04, 0x02, 0x08, 0x0C, 0x07, 0x08, 0x0F, 0x03, 0x08, 0x04, 0x0E, 0x0B, 0x13, 0x05, 
	0x08, 0x01, 0x07, 0x0A, 0x04, 0x02, 0x00, 0x0C, 0x13, 0x0E, 0x03, 0x13, 0x09, 0x0B, 0x09, 0x0B, 
	0x00, 0x11, 0x0F, 0x07, 0x09, 0x04, 0x00, 0x06, 0x0D, 0x0F, 0x01, 0x06, 0x10, 0x0D, 0x0A, 0x03
};
//////////////////////////////////////////////////
const char * capszGrades[20] = 
{
	"9", 
	"8", 
	"7", 
	"6", 
	"5", 
	"4", 
	"3", 
	"2", 
	"1", 
	"S1", 
	"S2", 
	"S3", 
	"S4", 
	"S5", 
	"S6", 
	"S7", 
	"S8", 
	"S9", 
	"M", 
	"GM"
};
//////////////////////////////////////////////////
const char * capszMedallions[4] = 
{
	"NONE", 
	"BRONZE", 
	"SILVER", 
	"GOLD"
};
//////////////////////////////////////////////////

int DecodeTAPScoreValidation(
		char * pstrNameABC,							// "KAN\0" or "CT \0"
		char * pstrScoreValidation,					// 16 character validation code to decode
		unsigned char * pbGameMode,					// [Optional pointer to store] LTAP_GAMEMODE_NORMAL or LTAP_GAMEMODE_MASTER
		unsigned short * pwVisibleGrade,			// [Optional pointer to store] LTAP_GRADE_M
		unsigned short * pwLevel,					// [Optional pointer to store] Max 300 for Normal mode or max 999 for other modes
		unsigned short * pwRankedLine,				// [Optional pointer to store] LTAP_RANKED_NONE
		unsigned long * pdwMasteringTime,			// [Optional pointer to store] TIME_TO_FRAMES(18, 31, 59) in all modes except Normal Mode
		unsigned long * pdwNormalModeScore,			// [Optional pointer to store] Score in Normal Mode only
		unsigned char * pbMedallionAllClear,		// [Optional pointer to store] LTAP_MEDALLION_NONE
		unsigned char * pbMedallionSectionTime, 	// [Optional pointer to store] LTAP_MEDALLION_NONE
		unsigned char * pbMedallionSkill, 			// [Optional pointer to store] LTAP_MEDALLION_NONE
		unsigned char * pbMedallionRecovery, 		// [Optional pointer to store] LTAP_MEDALLION_NONE
		unsigned char * pbMedallionRotation, 		// [Optional pointer to store] LTAP_MEDALLION_NONE
		unsigned char * pbMedallionCombo) 			// [Optional pointer to store] LTAP_MEDALLION_NONE
{
	char strComputeValidation[17] = {'\0'};
	char strScoreValidation[17] = {'\0'};
	char modulus;
	unsigned int i, j;
	unsigned short wGameMode;

	unsigned short wTheRand, wTheChecksum, wChecksum = 0, wStartOffset = 0;

	// Checksum the name (this will be used as a start offset of the code)
	for (i = 0; i < 3; i++)
	{
		wStartOffset += pstrNameABC[i];
	}

	//
	// Third pass reverse:
	// Convert code back to data with the correct offset
	//

	// We need a second copy of the code while we are decoding
	for (i = 0; i < 16; i++)
	{
		strScoreValidation[i] = pstrScoreValidation[i];
	}

	// Shift to correct start offset
	for (i = 0; i < 16; i++)
	{
		strComputeValidation[i] = strScoreValidation[(i - (wStartOffset & 0xF)) & 0xF];
	}

	// Convert validation code back to data where data in MOD 20 (between 0 and 19). This is only the first step because the origina data has been scrambled 
	// between 0 and 39!
	for (i = 0; i < 16; i++)
	{
		// The data
		modulus = strComputeValidation[i];

		for (j = 0; j < (unsigned int)sizeof(caValidationLetters); j++)
		{
			if (j == 0)
			{
				modulus -= caValidationLetters[0];
				strComputeValidation[i] = 0;
			}
			else
			{
				modulus -= (caValidationLetters[j] - caValidationLetters[j - 1]);
				++strComputeValidation[i];
			}

			if (modulus <= 0)
			{
				break;
			}
		}
	}

	//
	// Second pass reverse:
	// Unscramble the data using the stored random number and the scramble table
	//

	// The data now has valid randomizer and checksum. We need the randomizer to find the scrambled number used
	wTheRand = ((strComputeValidation[15] & 0xF) | ((strComputeValidation[14] & 0x1) << 4));

	// It is easier to have the scrambled numbers ready in an array for the computation below, just because we have memory for it
	for (i = 0; i <= 11; i++)
	{
		// Byte 0 to and including 11
		strScoreValidation[i] = caScramble[wTheRand];

		// Offset the randomizer within the scramble table
		wTheRand = ((wTheRand + 1) % (signed short)sizeof (caScramble));
	}

	// Unscramble the data. The scramble numbers are between 0 and 19, and were added before the modulo 20 operation, so we know the result 
	// value is somewhere between 0 and 39, so we start at 20 and if the result is above or equal to 20, we subtract 20 to get the real 
	// data, else it is the real data already
	for (i = 0; i <= 11; i++)
	{
		strComputeValidation[i] += 20;
		strComputeValidation[i] -= strScoreValidation[i];

		if (strComputeValidation[i] >= 20)
		{
			strComputeValidation[i] -= 20;
		}
	}

	//
	// First pass reverse:
	// Validate the data using the stored checksum, and if the data was valid, extract what was requested from the caller
	//

	// The data now has valid randomizer and checksum. We need the checksum to validate the data
	wTheChecksum = ((strComputeValidation[13] & 0xF) | ((strComputeValidation[12] & 0xF) << 4));

	// The checksum included the start offset, so we subtract it here, and mask the value to keep it binary compatible with our test below
	wTheChecksum -= wStartOffset;
	wTheChecksum &= 0xFF;

	// Time to calculate our current checksum
	for (i = 0; i <= 11; i++)
	{
		// Byte 0 to and including 11
		wChecksum += strComputeValidation[i];
	}
	for (i = 14; i <= 15; i++)
	{
		// Byte 14 to and including 15
		wChecksum += strComputeValidation[i];
	}

	//
	// At this point, if the checksum is not matching the stored checksum, either the code is invalid or 
	// the name was not correct. The name needs to be exactly the same as used while the code was computed
	//

	if (wTheChecksum != (wChecksum & 0xFF))
	{
		// Oh no, the code or the name was invalid or didn't match their original counterparts
		return 0;
	}

	//
	// We made it, well the checksum is not THAT GOOD, so it could still end up with incorrect data if unlucky. Lets extract the data
	//

	// Decode game mode. We need this one to decide if it is score or time to decode later
	wGameMode = (unsigned short)((strComputeValidation[0] >> 2) & 0x3);

	if (pbGameMode != NULL)
	{
		*pbGameMode = (unsigned char)wGameMode;
	}

	if (pwVisibleGrade != NULL)
	{
		// Decode visible grade (visible means shown in the HUD)
		*pwVisibleGrade = ((unsigned short)((strComputeValidation[1] >> 1) & 0x7) | ((unsigned short)(strComputeValidation[0] & 0x3) << 3));
	}

	if (pwLevel != NULL)
	{
		// Decode level
		*pwLevel = ((unsigned short)((strComputeValidation[4] >> 3) & 0x1) | 
					((unsigned short)(strComputeValidation[3] & 0xF) << 1) | 
					((unsigned short)(strComputeValidation[2] & 0xF) << 5) | 
					((unsigned short)(strComputeValidation[1] & 0x1) << 9));
	}

	if (pwRankedLine != NULL)
	{
		// Decode ranked line
		*pwRankedLine = (unsigned short)((strComputeValidation[4] >> 1) & 0x3);

		//
		// WARNING: If ranked line is > 2 then the code was messed up!
		//
	}

	// Decode score
	if (wGameMode == LTAP_GAMEMODE_NORMAL)
	{
		if (pdwNormalModeScore != NULL)
		{
			*pdwNormalModeScore = ((unsigned long)((strComputeValidation[9] >> 1) & 0x7) | 
								   ((unsigned long)(strComputeValidation[8] & 0xF) << 3) | 
								   ((unsigned long)(strComputeValidation[7] & 0xF) << 7) | 
								   ((unsigned long)(strComputeValidation[6] & 0xF) << 11) | 
								   ((unsigned long)(strComputeValidation[5] & 0xF) << 15) | 
								   ((unsigned long)(strComputeValidation[4] & 0x1) << 19));
		}

		if (pdwMasteringTime != NULL)
		{
			*pdwMasteringTime = 0;
		}
	}
	// Decode time
	else
	{
		if (pdwMasteringTime != NULL)
		{
			*pdwMasteringTime = ((unsigned long)((strComputeValidation[9] >> 1) & 0x7) | 
								 ((unsigned long)(strComputeValidation[8] & 0xF) << 3) | 
								 ((unsigned long)(strComputeValidation[7] & 0xF) << 7) | 
								 ((unsigned long)(strComputeValidation[6] & 0xF) << 11) | 
								 ((unsigned long)(strComputeValidation[5] & 0xF) << 15) | 
								 ((unsigned long)(strComputeValidation[4] & 0x1) << 19));
		}

		if (pdwNormalModeScore != NULL)
		{
			*pdwNormalModeScore = 0;
		}
	}

	if (pbMedallionAllClear != NULL)
	{
		// Decode [AC] medallion
		*pbMedallionAllClear = ((unsigned char)((strComputeValidation[10] >> 3) & 0x1) | 
								((unsigned char)(strComputeValidation[9] & 0x1) << 1));
	}

	if (pbMedallionSectionTime != NULL)	// Thanks WakiMiko for fixing this. See: http://www.tetrisconcept.net/threads/tap-hiscore-game-over-code-source-code.2483/#post-56705
	{
		// Decode [ST] medallion
		*pbMedallionSectionTime = (unsigned char)((strComputeValidation[10] >> 1) & 0x3);
	}

	if (pbMedallionSkill != NULL)
	{
		// Decode [SK] medallion
		*pbMedallionSkill = ((unsigned char)((strComputeValidation[11] >> 3) & 0x1) | 
							 ((unsigned char)(strComputeValidation[10] & 0x1) << 1));
	}

	if (pbMedallionRecovery != NULL)
	{
		// Decode [RE] medallion
		*pbMedallionRecovery = (unsigned char)((strComputeValidation[11] >> 1) & 0x3);
	}

	if (pbMedallionRotation != NULL)
	{
		// Decode [RO] medallion
		*pbMedallionRotation = ((unsigned char)((strComputeValidation[14] >> 3) & 0x1) | 
								((unsigned char)(strComputeValidation[11] & 0x1) << 1));
	}

	if (pbMedallionCombo != NULL)
	{
		// Decode [C0] medallion
		*pbMedallionCombo = (unsigned char)((strComputeValidation[14] >> 1) & 0x3);
	}

	// And that's it! Return 1 for success
	return 1;
}
