//
// The textfiles module takes care of all loading and parsing of the textfiles ORDER and SCORE.
//

#include "Textfiles.h"

#include "HiscoreOrder.h"
#include "ScoreValidationDec.h"
#include "Account.h"

#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

void LoadOrderTxt(char * pszFilename)
{
	// Default the order in case of file read/write error
	bCaptionOrder[0] = INDEX_ORDER_GRADE;
	bCaptionOrder[1] = INDEX_ORDER_SCORE;
	bCaptionOrder[2] = INDEX_ORDER_LEVEL;
	bCaptionOrder[3] = INDEX_ORDER_TIME;
	bCaptionOrder[4] = INDEX_ORDER_MEDAL;

	bMedallionOrder[0] = INDEX_ORDER_AC;
	bMedallionOrder[1] = INDEX_ORDER_ST;
	bMedallionOrder[2] = INDEX_ORDER_SK;
	bMedallionOrder[3] = INDEX_ORDER_RE;
	bMedallionOrder[4] = INDEX_ORDER_RO;
	bMedallionOrder[5] = INDEX_ORDER_CO;

	FILE * pFile = fopen(pszFilename, "rb");

	if (pFile == NULL)
	{
		if ((pFile = fopen(pszFilename, "wb")) != NULL)
		{
			fprintf(pFile, 
				"\r\n# This is the default ranking order:\r\n%s %s %s %s %s\r\n\r\n# We will rank medallions by default unless the line below is removed:\r\n%s %s %s %s %s %s", 
				aszOrderCaption[bCaptionOrder[0]],		// Grade
				aszOrderCaption[bCaptionOrder[1]],		// Score
				aszOrderCaption[bCaptionOrder[2]],		// Level
				aszOrderCaption[bCaptionOrder[3]],		// Time
				aszOrderCaption[bCaptionOrder[4]],		// Medal

				aszOrderMedal[bMedallionOrder[0]], 
				aszOrderMedal[bMedallionOrder[1]], 
				aszOrderMedal[bMedallionOrder[2]], 
				aszOrderMedal[bMedallionOrder[3]], 
				aszOrderMedal[bMedallionOrder[4]], 
				aszOrderMedal[bMedallionOrder[5]]);

			fclose(pFile);
		}

		return;
	}

	// Let's parse two lines to get the order of hiscore priority
	char szLine[128];
	char * pszToken;
	unsigned int i;
	unsigned char bDone;

	char ch = ' ';

	while (fgets(szLine, sizeof (szLine), pFile) != NULL)
	{
		// Make the whole line capitalized so we can find the words
		for (i = 0; i < strlen(szLine); i++)
		{
			if (ch == ' ' && szLine[i] != ch && szLine[i] != '\t')
			{
				ch = szLine[i];

				// If we encounter a hash symbol as the first character, it is a line comment so skip 
				// the entire line and try the next one
				if (ch == '#')
				{
					break;
				}
			}

			szLine[i] = toupper(szLine[i]);
		}

		if (ch == '#' || 
			ch == '\r' || 
			ch == '\n')
		{
			ch = ' ';
			continue;
		}

		i = 0;
		bDone = 0x0;

		pszToken = strtok(szLine, " \t");

		while (pszToken != NULL)
		{
			// Test for "GRADE"
			if (strncmp(aszOrderCaption[INDEX_ORDER_GRADE], pszToken, strlen(aszOrderCaption[INDEX_ORDER_GRADE])) == 0 && 
				!(bDone & 0x1))
			{
				bCaptionOrder[i] = INDEX_ORDER_GRADE;
				++i;
				bDone |= 0x1;
			}
			// Test for "TIME"
			else if (strncmp(aszOrderCaption[INDEX_ORDER_TIME], pszToken, strlen(aszOrderCaption[INDEX_ORDER_TIME])) == 0 && 
					 !(bDone & 0x2))
			{
				bCaptionOrder[i] = INDEX_ORDER_TIME;
				++i;
				bDone |= 0x2;
			}
			// Test for "SCORE"
			else if (strncmp(aszOrderCaption[INDEX_ORDER_SCORE], pszToken, strlen(aszOrderCaption[INDEX_ORDER_SCORE])) == 0 && 
					 !(bDone & 0x4))
			{
				bCaptionOrder[i] = INDEX_ORDER_SCORE;
				++i;
				bDone |= 0x4;
			}
			// Test for "LEVEL"
			else if (strncmp(aszOrderCaption[INDEX_ORDER_LEVEL], pszToken, strlen(aszOrderCaption[INDEX_ORDER_LEVEL])) == 0 && 
					 !(bDone & 0x8))
			{
				bCaptionOrder[i] = INDEX_ORDER_LEVEL;
				++i;
				bDone |= 0x8;
			}
			// Test for "MEDAL"
			else if (strncmp(aszOrderCaption[INDEX_ORDER_MEDAL], pszToken, strlen(aszOrderCaption[INDEX_ORDER_MEDAL])) == 0 && 
					 !(bDone & 0x10))
			{
				bCaptionOrder[i] = INDEX_ORDER_MEDAL;
				++i;
				bDone |= 0x10;
			}

			if (i >= 5)
			{
				break;
			}

			pszToken = strtok(NULL, " \t");
		}

		// So whatever was left should be added in the default order
		while ((bDone & 0x1F) != 0x1F)
		{
			// Add "GRADE" if not previously added
			if (!(bDone & 0x1))
			{
				bCaptionOrder[i] = INDEX_ORDER_GRADE;
				++i;
				bDone |= 0x1;
			}
			// Add "TIME" if not previously added
			else if (!(bDone & 0x2))
			{
				bCaptionOrder[i] = INDEX_ORDER_TIME;
				++i;
				bDone |= 0x2;
			}
			// Add "SCORE" if not previously added
			else if (!(bDone & 0x4))
			{
				bCaptionOrder[i] = INDEX_ORDER_SCORE;
				++i;
				bDone |= 0x4;
			}
			// Add "LEVEL" if not previously added
			else if (!(bDone & 0x8))
			{
				bCaptionOrder[i] = INDEX_ORDER_LEVEL;
				++i;
				bDone |= 0x8;
			}
			// Add "MEDAL" if not previously added
			else if (!(bDone & 0x10))
			{
				bCaptionOrder[i] = INDEX_ORDER_MEDAL;
				++i;
				bDone |= 0x10;
			}
		}

		// Always exit the while loop
		break;
	}

	// In case we don't find a line of medallions in the text file, we will not rank medallions
	FDoNotRankMedallions = true;

	ch = ' ';

	while (fgets(szLine, sizeof (szLine), pFile) != NULL)
	{
		// Make the whole line capitalized so we can find the words
		for (i = 0; i < strlen(szLine); i++)
		{
			if (ch == ' ' && szLine[i] != ch && szLine[i] != '\t')
			{
				ch = szLine[i];

				// If we encounter a hash symbol as the first character, it is a line comment so skip 
				// the entire line and try the next one
				if (ch == '#')
				{
					break;
				}
			}

			szLine[i] = toupper(szLine[i]);
		}

		if (ch == '#' || 
			ch == '\r' || 
			ch == '\n')
		{
			ch = ' ';
			continue;
		}

		FDoNotRankMedallions = false;

		i = 0;
		bDone = 0x0;

		pszToken = strtok(szLine, " \t");

		while (pszToken != NULL)
		{
			// Test for "AC"
			if (strncmp(aszOrderMedal[INDEX_ORDER_AC], pszToken, strlen(aszOrderMedal[INDEX_ORDER_AC])) == 0 && 
				!(bDone & 0x1))
			{
				bMedallionOrder[i] = INDEX_ORDER_AC;
				++i;
				bDone |= 0x1;
			}
			// Test for "ST"
			else if (strncmp(aszOrderMedal[INDEX_ORDER_ST], pszToken, strlen(aszOrderMedal[INDEX_ORDER_ST])) == 0 && 
					 !(bDone & 0x2))
			{
				bMedallionOrder[i] = INDEX_ORDER_ST;
				++i;
				bDone |= 0x2;
			}
			// Test for "SK"
			else if (strncmp(aszOrderMedal[INDEX_ORDER_SK], pszToken, strlen(aszOrderMedal[INDEX_ORDER_SK])) == 0 && 
					 !(bDone & 0x4))
			{
				bMedallionOrder[i] = INDEX_ORDER_SK;
				++i;
				bDone |= 0x4;
			}
			// Test for "RE"
			else if (strncmp(aszOrderMedal[INDEX_ORDER_RE], pszToken, strlen(aszOrderMedal[INDEX_ORDER_RE])) == 0 && 
					 !(bDone & 0x8))
			{
				bMedallionOrder[i] = INDEX_ORDER_RE;
				++i;
				bDone |= 0x8;
			}
			// Test for "RO"
			else if (strncmp(aszOrderMedal[INDEX_ORDER_RO], pszToken, strlen(aszOrderMedal[INDEX_ORDER_RO])) == 0 && 
					 !(bDone & 0x10))
			{
				bMedallionOrder[i] = INDEX_ORDER_RO;
				++i;
				bDone |= 0x10;
			}
			// Test for "CO"
			else if (strncmp(aszOrderMedal[INDEX_ORDER_CO], pszToken, strlen(aszOrderMedal[INDEX_ORDER_CO])) == 0 && 
					 !(bDone & 0x20))
			{
				bMedallionOrder[i] = INDEX_ORDER_CO;
				++i;
				bDone |= 0x20;
			}

			if (i >= 6)
			{
				break;
			}

			pszToken = strtok(NULL, " \t");
		}

		if (bDone == 0x0)
		{
			// No, noone wanted medallions order so we must obay
			FDoNotRankMedallions = true;

			fclose(pFile);

			return;
		}

		// So whatever was left should be added in the default order
		while ((bDone & 0x3F) != 0x3F)
		{
			// Add "AC" if not previously added
			if (!(bDone & 0x1))
			{
				bMedallionOrder[i] = INDEX_ORDER_AC;
				++i;
				bDone |= 0x1;
			}
			// Add "ST" if not previously added
			else if (!(bDone & 0x2))
			{
				bMedallionOrder[i] = INDEX_ORDER_ST;
				++i;
				bDone |= 0x2;
			}
			// Add "SK" if not previously added
			else if (!(bDone & 0x4))
			{
				bMedallionOrder[i] = INDEX_ORDER_SK;
				++i;
				bDone |= 0x4;
			}
			// Add "RE" if not previously added
			else if (!(bDone & 0x8))
			{
				bMedallionOrder[i] = INDEX_ORDER_RE;
				++i;
				bDone |= 0x8;
			}
			// Add "RO" if not previously added
			else if (!(bDone & 0x10))
			{
				bMedallionOrder[i] = INDEX_ORDER_RO;
				++i;
				bDone |= 0x10;
			}
			// Add "CO" if not previously added
			else if (!(bDone & 0x20))
			{
				bMedallionOrder[i] = INDEX_ORDER_CO;
				++i;
				bDone |= 0x20;
			}
		}

		// Always exit the while loop
		break;
	}

	fclose(pFile);
}

static void ErrorOutAndFreeAccount(Account * pAccount, unsigned int uParsedLine)
{
	printf("There was an error in your score table file on line %d:\r\nName: \"", uParsedLine);

	for (uParsedLine = 0; uParsedLine < 3; uParsedLine++)
	{
		putc(pAccount->aName[uParsedLine], stdout);
	}

	printf("\"\r\nCode: ");

	for (uParsedLine = 0; uParsedLine < 16; uParsedLine++)
	{
		putc(pAccount->aCode[uParsedLine], stdout);
	}

	printf("\r\n\r\n");

	free(pAccount);
}

bool LoadScoreTxt(char * pszFilename)
{
	FILE * pFile = fopen(pszFilename, "rb+");
	bool FNoErrors = true;

	if (pFile == NULL)
	{
		if ((pFile = fopen(pszFilename, "wb")) != NULL)
		{
			//
			// TO-DO: Add default TAP hiscore with their exact level and hidden stuff! It will be a good reference!
			//

			fclose(pFile);
		}

		return true;
	}

	// I will manually parse each line because I find it easier to control the length and format of things
	char szLine[256];
	unsigned int i, j, k, uLine = 0;

	Account * pAccountNew;

	while (uLine++, fgets(szLine, sizeof (szLine), pFile) != NULL)
	{
		char ch = ' ';

		// Test for line comment or empty line
		for (i = 0; i < strlen(szLine); i++)
		{
			if (szLine[i] != ch && szLine[i] != '\t')
			{
				ch = szLine[i];
				break;
			}
		}

		// If we encounter a hash symbol as the first character, it is a line comment so skip 
		// the entire line and try the next one
		if (ch == '#' || 
			ch == '\r' || 
			ch == '\n')
		{
			continue;
		}

		// We allocate the account first, so we can use its storage for testing. If the line we read in is incomplete or invalid, we will have to 
		// deallocate the memory before we continue to the next line. However if the line is valid, the allocation will live on until CleanupAccounts() 
		// cleans it up and deallocates the memory. Either way, pAccountNew is an imporant pointer until the next iteration of the while loop
		if ((pAccountNew = (Account *)malloc(sizeof (Account))) == NULL)
		{
			// Out of memory, bail out
			break;
		}

		// Because we got an error output we will clear the data with whitespaces. See function ErrorOutAndFreeAccount() which can be 
		// called before we have a valid name or code, and those are not NULL-terminated strings!
		{
			for (j = 0; j < 3; j++)
			{
				pAccountNew->aName[j] = ' ';
			}

			for (j = 0; j < 16; j++)
			{
				pAccountNew->aCode[j] = ' ';
			}
		}

		// We will parse through 127 characters in hope to find our data
		i = 0;

		// Find the first non-whitespace/tab character
		while (i < 127 && 
			   szLine[i] == ' ' || 
			   szLine[i] == '\t')
		{
			++i;
		}

		if (i == 127)
		{
			// No more characters to read, cleanup and do the next line read iteration
			ErrorOutAndFreeAccount(pAccountNew, uLine);
			FNoErrors = false;
			continue;
		}

		// Now, we need to read in the 3 character name. Because we allow names less than 3, we must stop when we find a whitespace character. Also names can begin 
		// with whitespaces so we must check for quote signs at beginning and end
		char chLastQuote = 0;

		for (j = 0; j < 3; j++)
		{
			if (j == 0)
			{
				if (chLastQuote == 0 && 
					szLine[i] == '\"' || 
					szLine[i] == '\'')
				{
					chLastQuote = szLine[i];

					if (++i >= 127)
					{
						// We are out of space. This entry is invalid
						break;
					}

					--j;

					continue;
				}
			}

			if (szLine[i] == '\t')
			{
				// Add whitespaces to the rest of the name and move on
				for (j; j < 3; j++)
				{
					pAccountNew->aName[j] = ' ';
				}

				break;
			}

			// Also, convert to a capital letter
			pAccountNew->aName[j] = toupper(szLine[i]);

			if (++i >= 127)
			{
				// We are out of space. This entry is invalid
				break;
			}
		}

		// We should end with the same quote, just simply because it is not a whitespace sequence and we 
		// don't want to promote bad formatting and invalid code reads below
		if (chLastQuote != 0)
		{
			// Find that darn quote or fail this entry
			for (j = 0; j < 2; j++)
			{
				if (szLine[i] == chLastQuote)
				{
					// Got it, continue on...
					chLastQuote = 0;
					j = 2;
				}

				if (++i >= 127)
				{
					// We are out of space. This entry is invalid
					break;
				}
			}

			if (chLastQuote != 0)
			{
				// No, this is bad formatting. Do a better job next time
				ErrorOutAndFreeAccount(pAccountNew, uLine);
				FNoErrors = false;
				continue;
			}
		}

		if (i == 127)
		{
			// No more characters to read, cleanup and do the next line read iteration
			ErrorOutAndFreeAccount(pAccountNew, uLine);
			FNoErrors = false;
			continue;
		}

		// Time to format the name. If some of the ascii characters are not supported, we will stop the name 
		// and insert whitespaces. It is pretty much a fail, because there is a very little chance the score 
		// validation code will work with an invalid name. But we will try any way
		for (j = 0; j < 3; j++)
		{
			// The is an optimization to test if a character is invalid at the end of the next foor loop. We 
			// use the MSB to tell if the character was modified, and if it was, it was validated!
			pAccountNew->aName[j] |= 0x80;

			// Validate the letter
			for (k = 1; k < (((unsigned int)sizeof (caHiscoreNameLetters)) - 2); k++)
			{
				if (((unsigned char)pAccountNew->aName[j] & ~0x80) == (unsigned char)caHiscoreNameLetters[k])
				{
					// Yet, validated
					pAccountNew->aName[j] = caHiscoreNameLetters[k];
					break;
				}
			}

			// Did we succeed?
			if (!(pAccountNew->aName[j] & 0x80))
			{
				// Yep, continue to the next letter
				continue;
			}
			else
			{
				// No, this ascii letter was invalid, and since we can't have whitespace in the middle of a name, we clear the 
				// rest of the characters and hope for the best
				for (j; j < 3; j++)
				{
					pAccountNew->aName[j] = ' ';
				}

				//
				// We will end the current for loop automatically because j is now 3
				//
			}
		}

		// Next up is the score validation code. We need the next letter that is not a whitespace or a tab
		while (i < 127 && 
			   szLine[i] == ' ' || 
			   szLine[i] == '\t')
		{
			++i;
		}

		if (i == 127)
		{
			// No more characters to read, cleanup and do the next line read iteration
			ErrorOutAndFreeAccount(pAccountNew, uLine);
			FNoErrors = false;
			continue;
		}

		// So the code needs to be 16 characters long. If we find a whitespace or a tab anywhere before that, it is an instant fail
		for (j = 0; j < 16; j++)
		{
			if (szLine[i] == ' ' || 
				szLine[i] == '\t')
			{
				// An easy way to fail is to set i to 127 and break
				i = 127;
				break;
			}

			// Also, convert to a capital letter
			pAccountNew->aCode[j] = toupper(szLine[i]);

			if (++i >= 127)
			{
				// We are out of space. This entry is invalid
				break;
			}
		}

		if (i == 127)
		{
			// No more characters to read, cleanup and do the next line read iteration
			ErrorOutAndFreeAccount(pAccountNew, uLine);
			FNoErrors = false;
			continue;
		}

		// Validate the code's characters. A code can only use 20 predefined easy-to-remember characters
		for (j = 0; j < 16; j++)
		{
			// The is an optimization to test if a character is invalid at the end of the next foor loop. We 
			// use the MSB to tell if the character was modified, and if it was, it was validated!
			pAccountNew->aCode[j] |= 0x80;

			// Validate the letter
			for (k = 0; k < ((unsigned int)sizeof (caValidationLetters)); k++)
			{
				if (((unsigned char)pAccountNew->aCode[j] & ~0x80) == (unsigned char)caValidationLetters[k])
				{
					// Yet, validated
					pAccountNew->aCode[j] = caValidationLetters[k];
					break;
				}
			}

			// Did we succeed?
			if (!(pAccountNew->aCode[j] & 0x80))
			{
				// Yep, continue to the next letter
				continue;
			}
			else
			{
				// No, instant fail!
				j = 16;
				i = 127;

				//
				// We will end the current for loop automatically because j is now 16
				//
			}
		}

		if (i == 127)
		{
			// No more characters to read, cleanup and do the next line read iteration
			ErrorOutAndFreeAccount(pAccountNew, uLine);
			FNoErrors = false;
			continue;
		}

		// Last we add any comment. A comment begins with a non-whitespace/tab character and ends at the line end. This is also were we allocate 
		// the comment characters for the account. It is very important that the pointer to this allocation is set to either the allocation or 
		// NULL as it will be our only constructor requirement!
		while (i < 127 && 
			   szLine[i] == ' ' || 
			   szLine[i] == '\t')
		{
			++i;
		}

		if (i == 127)
		{
			// Comment is optional, so we will not fail. We MUST set the pointer to NULL though!
			pAccountNew->pszComment = NULL;
		}
		else
		{
			// First, we need to know the size of the comment. While we are doing this we need to remember where the comment 
			// started in the line buffer
			j = i;
			k = 0;

			while (i < 127)
			{
				if (szLine[i] == '\r' || 
					szLine[i] == '\n' || 
					szLine[i] == '\0')
				{
					break;
				}

				++k;
				++i;
			}

			// If there were no characters in the comment, we MUST set the pointer to NULL
			if (k == 0)
			{
				pAccountNew->pszComment = NULL;
			}
			else
			{
				// First, the allocated comment needs to be NULL-terminated so add 1 more character to the allocation. 
				// If memory fails us, we will still continue but the comment will be discarded
				if ((pAccountNew->pszComment = (char *)malloc(k + 1)) != NULL)
				{
					// NULL-terminate it
					pAccountNew->pszComment[k] = '\0';

					// String copy the comment, and use k number of characters
					strncpy(pAccountNew->pszComment, &szLine[j], k);
				}
			}

			// Okay, it is time. We have the account ready for decoding. One thing to note is the name and code as they are expected to be 
			// NULL-terminated, but my decode function never uses the last NULL byte, so we can get away with this
			if (DecodeTAPScoreValidation(
					pAccountNew->aName, 
					pAccountNew->aCode, 
					&pAccountNew->bMode, 
					&pAccountNew->wGrade, 
					&pAccountNew->wLevel, 
					&pAccountNew->wRankedLine, 
					&pAccountNew->dwTime, 
					&pAccountNew->dwScore, 
					&pAccountNew->abMedallions[INDEX_ORDER_AC], 
					&pAccountNew->abMedallions[INDEX_ORDER_ST], 
					&pAccountNew->abMedallions[INDEX_ORDER_SK], 
					&pAccountNew->abMedallions[INDEX_ORDER_RE], 
					&pAccountNew->abMedallions[INDEX_ORDER_RO], 
					&pAccountNew->abMedallions[INDEX_ORDER_CO]) != 1)
			{
				// Invalid name/code combination. This account is bogus, so throw away and move on
				ErrorOutAndFreeAccount(pAccountNew, uLine);
				FNoErrors = false;
				continue;
			}

			// At this point in time, we have the rights to apply rules to the hiscore entry since the score validation code 
			// contains some information not valid in certain game modes. Mainly clearing the grade from interfering with 
			// Normal Mode, Plus Mode, and some of the Death Mode play must be done here, as well as ranking Normal Mode 
			// without the green or orange line (green line is applied to level 300 games as visual only)
			if (pAccountNew->bMode == LTAP_GAMEMODE_NORMAL || 
				pAccountNew->bMode == LTAP_GAMEMODE_PLUS || 
				pAccountNew->bMode == LTAP_GAMEMODE_DEATH)
			{
				if ((pAccountNew->bMode == LTAP_GAMEMODE_DEATH && 
					 pAccountNew->wGrade < LTAP_GRADE_M) || 
					pAccountNew->bMode != LTAP_GAMEMODE_DEATH)
				{
					// Clear the grade so that it doesn't interfere with the score order
					pAccountNew->wGrade = LTAP_GRADE_9;
				}

				if (pAccountNew->bMode == LTAP_GAMEMODE_NORMAL)
				{
					// If ranked with time first, the header will be added last in the Normal Mode section unless 
					// we adjust the time
					if (bCaptionOrder[0] == INDEX_ORDER_TIME)
					{
						// Bogus time for ranking ordering
						pAccountNew->dwTime = (0 + 4);
					}
				}

				// For Normal Mode, Plus Mode and Death Mode, staff roll can be "skipped" by holding the start button. That means 
				// the hiscore should always count finished games as green line. Normal Mode does it this way even when the Score 
				// Validation Code says orangle line!
				pAccountNew->wRankedLine = LTAP_RANKED_NONE;
			}

			// The function AddAccount() helps us adding the account to the right hisocre mode group
			if (AddAccount(pAccountNew) == false)
			{
				// Something went wrong, so deallocate and move on
				ErrorOutAndFreeAccount(pAccountNew, uLine);
				FNoErrors = false;
				continue;
			}

			//
			// Account was added, let's move on the the next one...
			//
		}
	}

	fclose(pFile);

	return FNoErrors;
}
